import json
from django.shortcuts import render, redirect, reverse, HttpResponseRedirect, HttpResponse
from django.views.generic import ListView, DetailView, CreateView
from django.db.models import Q
from django.http import JsonResponse
import json

from .models import News, WpEsiNews, WpEsiNewsEntity, WpEsiMedia, WpEsiEntity


class ListNewsView(ListView):
    template_name = 'list_news.html'

    def get(self, request, *args, **kwargs):
        media = request.GET.get('media')
        from_date = request.GET.get('from_date')
        to_date = request.GET.get('to_date')
        search = request.GET.get('search')
        print(from_date, to_date)
        news = News.objects.filter(is_accepted=False).order_by('-id')#.order_by('record_date')
        if search:
            news = news.filter(Q(title__icontains=search) | Q(description__icontains=search))
        if from_date:
            news = news.filter(record_date__gte=from_date)
        if to_date:
            news = news.filter(record_date__lte=to_date)
        if media:
            news = self._filter_by_media(news, media)
        media = WpEsiMedia.objects.all()
        return render(request, self.template_name, {'news': news,
                                                    'media': media})

    def post(self, request, *args, **kwargs):
        if request.is_ajax():
            data = json.loads(request.body.decode('utf-8'))
            print(data)
            type_request = data['type']
            ids = data['id']
            if type_request in ['delete', 'deleteAll']:
                self._delete(ids)
            if type_request == 'edit':
                return HttpResponse({'ids': ids})
        return HttpResponse(200)

    def _delete(self, id):
        if isinstance(id, list):
            News.objects.filter(id__in=id).update(is_accepted=True)
        else:
            News.objects.filter(id=id).update(is_accepted=True)

    def _filter_by_media(self, news, media):
        if media == 'All':
            return news
        else:
            m = WpEsiMedia.objects.get(name=media)
            news = news.filter(media_id=m.pk)
            return news


class EditNewsView(DetailView):
    template_name = 'edit_news.html'

    def get(self, request, *args, **kwargs):
        news = News.objects.get(id=kwargs.get('id'))
        media = WpEsiMedia.objects.all()
        return render(request, self.template_name, {'news': news,
                                                    'media': media})

    def post(self, request, *args, **kwargs):
        radar = request.POST.get('radar')
        print(radar)

        return HttpResponseRedirect('/news')


class ViewNewsView(DetailView):
    template_name = 'view_news.html'

    def get(self, request, *args, **kwargs):
        news_id = kwargs.get('id')
        news = News.objects.get(id=news_id)

        return render(request, self.template_name, {'news': news})

    def post(self, request, *args, **kwargs):
        delete_news = request.POST.get('delete')
        export_news = request.POST.get('export')

        if delete_news:
            News.objects.filter(id=delete_news).update(is_accepted=True)
            return HttpResponseRedirect(reverse('view_news', kwargs={'id': delete_news}))

        if export_news:
            news = News.objects.get(id=export_news)
            esi = WpEsiNews.objects.create(
                title=news.title,
                description=news.description,
                url=news.url,
                media_id=news.media_id,
                type_id=news.type_id,
                region_id=news.region_id,
                post_id=news.post_id,
                publish_date=news.publish_date,
                record_date=news.record_date
            )
            esi_entity = WpEsiNewsEntity.objects.create(
                news_id=esi.id,
                entity_id=news.company_id
            )
            return HttpResponseRedirect(reverse('view_news', kwargs={'id': export_news}))


def radar(request, id):
    d = [{"text": "Algeria", "value" : "Algeria"},
  {"text": "Angola", "value" : "Angola"},
  {"text": "Angola", "value" : "as"},
  {"text": "Angola", "value" : "dsda"},
  {"text": "Angola", "value" : "aasdadsd"}]
    return JsonResponse(d, safe=False)